#define TRACE_NAME "MPITransfer"
#include "artdaq-mpich-plugin/TransferPlugins/MPITransfer.hh"
#include <algorithm>

#include "canvas/Utilities/Exception.h"
#include "cetlib_except/exception.h"

#include "artdaq-core/Data/Fragment.hh"


/*
  Protocol: want to do a send for each request object, then wait for for
  pending requests to complete, followed by a reset to allow another set
  of sends to be completed.

  This needs to be separated into a thing for sending and a thing for receiving.
  There probably needs to be a common class that both use.
*/

#define MPI_TAG_HEADER 0x8E // 142
#define MPI_TAG_DATA 0xDA   // 218
#define USE_RECV 1

std::mutex artdaq::MPITransfer::mpi_mutex_;

artdaq::MPITransfer::MPITransfer(fhicl::ParameterSet pset, Role role)
	: TransferInterface(pset, role)
	, reqs_(2 * buffer_count_, MPI_REQUEST_NULL)
	, payload_(buffer_count_)
	, pos_()
{
	TLOG(TLVL_TRACE) << GetTraceName() << " construction: "
		<< "source rank " << source_rank() << ", "
		<< "destination rank " << destination_rank() << ", "
		<< buffer_count_ << " buffers. ";

	if (buffer_count_ == 0)
	{
		throw art::Exception(art::errors::Configuration, "MPITransfer: ")
			<< "No buffers configured.";
	}
}

artdaq::MPITransfer::
~MPITransfer()
{
	TLOG(TLVL_TRACE) << GetTraceName() << ": ~MPITransfer: BEGIN";
	TLOG(TLVL_TRACE) << GetTraceName() << ": ~MPITransfer: Collecting requests that need to be waited on";
	std::vector<MPI_Request> reqs;
	for (size_t ii = 0; ii < reqs_.size(); ++ii)
	{
		if (reqs_[ii] != MPI_REQUEST_NULL)
		{
			reqs.push_back(reqs_[ii]);
		}
	}
	if (reqs.size() > 0)
	{
		TLOG(TLVL_TRACE) << GetTraceName() << ": ~MPITransfer: Waiting on " << reqs.size() << " reqs.";
		MPI_Waitall(reqs.size(), &reqs[0], MPI_STATUSES_IGNORE);
	}
	/*
	  TLOG_ARB(TLVL_VERBOSE, "MPITransfer") << uniqueLabel() << " ::~MPITransfer: Entering Barrier");
	MPI_Barrier(MPI_COMM_WORLD);
	TLOG_ARB(TLVL_VERBOSE, "MPITransfer") << uniqueLabel() << " ::~MPITransfer: Done with Barrier");*/
	TLOG(TLVL_TRACE) << GetTraceName() << ": ~MPITransfer: DONE";
}

artdaq::TransferInterface::CopyStatus
artdaq::MPITransfer::
copyFragment(Fragment& frag, size_t send_timeout_usec)
{
	return sendFragment(std::move(frag), send_timeout_usec);
}

artdaq::TransferInterface::CopyStatus
artdaq::MPITransfer::
moveFragment(Fragment&& frag)
{
	return sendFragment(std::move(frag), 0);
}

artdaq::TransferInterface::CopyStatus
artdaq::MPITransfer::
sendFragment(Fragment&& frag, size_t send_timeout_usec)
{
	if (frag.dataSize() > max_fragment_size_words_)
	{
		TLOG(TLVL_WARNING) << GetTraceName() << " Fragment has size (" << frag.dataSize() << ") larger than max_fragment_size_words_ (" << max_fragment_size_words_ << ")."
			<< " Total buffer space is: " << max_fragment_size_words_ * buffer_count_ << " words. Multiple over-size Fragments will exhaust the buffer!";
	}

	auto start_time = std::chrono::steady_clock::now();

	TLOG(5) << GetTraceName() << ": moveFragment: Finding available send slot, send_timeout_usec=" << send_timeout_usec;
	auto req_idx = findAvailable();
	auto counter = 0;
	while (req_idx == RECV_TIMEOUT && (send_timeout_usec == 0 || TimeUtils::GetElapsedTimeMicroseconds(start_time) < send_timeout_usec))
	{
		usleep(1000);
		req_idx = findAvailable();
		counter++;
		if (counter % 1000 == 0)
		{
			TLOG(TLVL_INFO) << GetTraceName() << " Rank " << source_rank() << " waiting for available buffer to " << destination_rank() << ". "
				<< "Waited " << TimeUtils::GetElapsedTimeMilliseconds(start_time) << " ms so far.";
		}

	}
	if (req_idx == TransferInterface::RECV_TIMEOUT)
	{
		TLOG(TLVL_WARNING) << GetTraceName() << ": sendFragment: No buffers available! Returning RECV_TIMEOUT!";
		return CopyStatus::kTimeout;
	}

	TLOG(5) << GetTraceName() << ": moveFragment send slot is " << req_idx;
	auto buffer_idx = req_idx / 2;
	TLOG(5) << GetTraceName() << ": moveFragment: Swapping in fragment to send to buffer " << buffer_idx;
	Fragment& curfrag = payload_[buffer_idx];
	curfrag = std::move(frag);

	TLOG(5) << GetTraceName() << ": moveFragment before send src=" << source_rank() << " dest=" << destination_rank() << " seqID=" << curfrag.sequenceID() << " type=" << curfrag.typeString() << " found_idx=" << req_idx;

	std::unique_lock<std::mutex> lk(mpi_mutex_);

	// 14-Sep-2015, KAB: we should consider MPI_Issend here (see below)...
	TLOG(5) << GetTraceName() << ": moveFragment: Using MPI_Isend";
	//Waits for the receiver to acknowledge header
	MPI_Issend(curfrag.headerAddress(), detail::RawFragmentHeader::num_words() * sizeof(RawDataType), MPI_BYTE, destination_rank(), MPI_TAG_HEADER, MPI_COMM_WORLD, &reqs_[req_idx]);

	auto sizeWrds = curfrag.size() - detail::RawFragmentHeader::num_words();
	auto offset = curfrag.headerAddress() + detail::RawFragmentHeader::num_words();
	MPI_Issend(offset, sizeWrds * sizeof(RawDataType), MPI_BYTE, destination_rank(), MPI_TAG_DATA, MPI_COMM_WORLD, &reqs_[req_idx + 1]);
	TLOG(5) << GetTraceName() << ": moveFragment COMPLETE";

	TLOG(11) << GetTraceName() << ": moveFragment COMPLETE: "
		<< " buffer_idx=" << buffer_idx
		<< " send_size=" << curfrag.size()
		<< " src=" << source_rank()
		<< " dest=" << destination_rank()
		<< " sequenceID=" << curfrag.sequenceID()
		<< " fragID=" << curfrag.fragmentID();
	return CopyStatus::kSuccess;
}

int artdaq::MPITransfer::receiveFragmentHeader(detail::RawFragmentHeader& header, size_t timeout_usec)
{
	TLOG(6) << GetTraceName() << ": receiveFragmentHeader entered tmo=" << timeout_usec << " us (ignored)";
	MPI_Status status;
	int wait_result = MPI_SUCCESS;

	MPI_Request req;
	{
		std::unique_lock<std::mutex> lk(mpi_mutex_);
		MPI_Irecv(&header, header.num_words() * sizeof(RawDataType), MPI_BYTE, source_rank(), MPI_TAG_HEADER, MPI_COMM_WORLD, &req);
	}
	//TLOG(TLVL_DEBUG) << GetTraceName() << ": receiveFragmentHeader: Start of receiveFragment" ;

	int flag;
	do {
		std::unique_lock<std::mutex> lk(mpi_mutex_);
		wait_result = MPI_Test(&req, &flag, &status);
		if (!flag) {
			usleep(1000);
			//TLOG_ARB(6) << uniqueLabel() << " MPITransfer::receiveFragmentHeader wait loop, flag=" << flag ;
		}
	} while (!flag);

	if (req != MPI_REQUEST_NULL)
	{
		TLOG(TLVL_ERROR) << GetTraceName() << " INTERNAL ERROR: req is not MPI_REQUEST_NULL in receiveFragmentHeader.";
		throw art::Exception(art::errors::LogicError, "MPITransfer: ") << "INTERNAL ERROR: req is not MPI_REQUEST_NULL in receiveFragmentHeader.";
	}
	//TLOG(TLVL_DEBUG) << GetTraceName() << " After testing/waiting res=" << wait_result ;
	TLOG(8) << GetTraceName() << ": receiveFragmentHeader recvd";


	{TLOG(8) << GetTraceName() << ": receiveFragmentHeader: " << my_rank
		<< " Wait_error=" << wait_result
		<< " status_error=" << status.MPI_ERROR
		<< " source=" << status.MPI_SOURCE
		<< " tag=" << status.MPI_TAG
		<< " Fragment_sequenceID=" << (uint64_t)header.sequence_id
		<< " Fragment_size=" << header.word_count
		<< " fragID=" << header.fragment_id;
	}
	char err_buffer[MPI_MAX_ERROR_STRING];
	int resultlen;
	switch (wait_result)
	{
	case MPI_SUCCESS:
		break;
	case MPI_ERR_IN_STATUS:
		MPI_Error_string(status.MPI_ERROR, err_buffer, &resultlen);
		TLOG(TLVL_ERROR) << GetTraceName()
			<< ": Waitany ERROR: " << err_buffer;
		break;
	default:
		MPI_Error_string(wait_result, err_buffer, &resultlen);
		TLOG(TLVL_ERROR) << GetTraceName()
			<< ": Waitany ERROR: " << err_buffer;
	}

	//TLOG_INFO) << GetTraceName() << " End of receiveFragment" ;
	return status.MPI_SOURCE;
}

int artdaq::MPITransfer::receiveFragmentData(RawDataType* destination, size_t wordCount)
{
	TLOG(6) << GetTraceName() << ": receiveFragmentData entered wordCount=" << wordCount;
	int wait_result = MPI_SUCCESS;
	MPI_Status status;

	MPI_Request req;
	{
		std::unique_lock<std::mutex> lk(mpi_mutex_);
		MPI_Irecv(destination, wordCount * sizeof(RawDataType), MPI_BYTE, source_rank(), MPI_TAG_DATA, MPI_COMM_WORLD, &req);
	}
	//TLOG(TLVL_DEBUG) << GetTraceName() << " Start of receiveFragment" ;

	int flag;
	do {
		std::unique_lock<std::mutex> lk(mpi_mutex_);
		wait_result = MPI_Test(&req, &flag, &status);
		if (!flag) {
			usleep(1000);
			//TLOG(6) << GetTraceName() << ": receiveFragmentData wait loop, flag=" << flag ;
		}
	} while (!flag);
	if (req != MPI_REQUEST_NULL)
	{
		TLOG(TLVL_ERROR) << GetTraceName() << " INTERNAL ERROR: req is not MPI_REQUEST_NULL in receiveFragmentData.";
		throw art::Exception(art::errors::LogicError, "MPITransfer: ") << "INTERNAL ERROR: req is not MPI_REQUEST_NULL in receiveFragmentData.";
	}

	//TLOG(TLVL_DEBUG) << GetTraceName() << " After testing/waiting res=" << wait_result ;
	TLOG(8) << GetTraceName() << ": receiveFragmentData recvd";


	char err_buffer[MPI_MAX_ERROR_STRING];
	int resultlen;
	switch (wait_result)
	{
	case MPI_SUCCESS:
		break;
	case MPI_ERR_IN_STATUS:
		MPI_Error_string(status.MPI_ERROR, err_buffer, &resultlen);
		TLOG(TLVL_ERROR) << GetTraceName()
			<< " MPITransfer: Waitany ERROR: " << err_buffer;
		break;
	default:
		MPI_Error_string(wait_result, err_buffer, &resultlen);
		TLOG(TLVL_ERROR) << GetTraceName()
			<< " MPITransfer: Waitany ERROR: " << err_buffer;
	}

	//TLOG_INFO) << GetTraceName() << " End of MPITransfer::receiveFragmentData" ;
	return status.MPI_SOURCE;
}

void
artdaq::MPITransfer::
cancelReq_(MPI_Request req) const
{
	if (req == MPI_REQUEST_NULL) return;

	TLOG(8) << GetTraceName() << ": Cancelling post";

	std::unique_lock<std::mutex> lk(mpi_mutex_);
	int result = MPI_Cancel(&req);
	if (result == MPI_SUCCESS)
	{
		MPI_Status status;
		MPI_Wait(&req, &status);
	}
	else
	{
		switch (result)
		{
		case MPI_ERR_REQUEST:
			throw art::Exception(art::errors::LogicError, "MPITransfer: ")
				<< "MPI_Cancel returned MPI_ERR_REQUEST.\n";
		case MPI_ERR_ARG:
			throw art::Exception(art::errors::LogicError, "MPITransfer: ")
				<< "MPI_Cancel returned MPI_ERR_ARG.\n";
		default:
			throw art::Exception(art::errors::LogicError, "MPITransfer: ")
				<< "MPI_Cancel returned unknown error code.\n";
		}
	}
}

int artdaq::MPITransfer::findAvailable()
{
	int use_me;
	int flag, flag2;
	size_t loops = 0;
	//TLOG_ARB(5) << uniqueLabel() <<  " findAvailable initial pos_=" << pos_ ;
	do
	{
		use_me = pos_;
		std::unique_lock<std::mutex> lk(mpi_mutex_);
		MPI_Test(&reqs_[use_me], &flag, MPI_STATUS_IGNORE);
		if (flag) {
			MPI_Test(&reqs_[use_me + 1], &flag2, MPI_STATUS_IGNORE);
		}
		pos_ = (pos_ + 2) % reqs_.size();
		++loops;
	} while (!flag2 && loops < buffer_count_);
	if (loops == buffer_count_) { return TransferInterface::RECV_TIMEOUT; }
	TLOG(5) << GetTraceName()  << " findAvailable returning use_me=" << use_me << " loops=" << loops;
	// pos_ is pointing at the next slot to check
	// use_me is pointing at the slot to use
	return use_me;
}

DEFINE_ARTDAQ_TRANSFER(artdaq::MPITransfer)
